# Generated by Django 4.2.21 on 2025-05-15 08:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('api', '0007_template_created_at_template_created_by_group_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='template',
            name='engine',
            field=models.CharField(choices=[('docx-template', 'Docx Template engine (https://github.com/elapouya/python-docx-template)'), ('xlsx-template', 'Xlsx Template engine (https://github.com/zhangyu836/python-xlsx-template)')], max_length=20),
        ),
    ]
